#include <utility>
#include <fstream>
#include <functional>
#include <iostream>
#include <ratio>

template<typename T>
struct escaper_t
{
    T t;
    bool b;
    operator bool() const { return b; }
};

template<typename T>
auto escaper(T&& t, bool cond){ return escaper_t<std::remove_reference_t<T>>{ std::forward<T>(t), cond }; }

//Minimal Pair implementation
template<typename L, typename R> struct Pair
{
    L l; R r;
};

template<typename L, typename R>
auto makePair(L&& l, R&& r){ return Pair<std::remove_reference_t<L>, std::remove_reference_t<R>>{ std::forward<L>(l), std::forward<R>(r) }; }

template<typename UF, typename FF, typename S, typename Z>
auto hylo( UF&& uf, FF&& ff, S&& seed, Z&& zero )
{
    auto ms = uf( std::forward<S>(seed) );
    auto tmp = std::forward<Z>(zero);
    while( ms )
    {
        tmp = ff( tmp, ms.t.l );
        ms = uf( ms.t.r );
    }
    return tmp;
}

template<typename F, typename dF, typename T>
auto NewtonIterator( F&& f, dF&& df, T&& start, T&& tolerance )
{
    return hylo(    [=](auto xn)
                    {
                        auto xnn = xn - f(xn)/df(xn);
                        return escaper( makePair(xnn, xnn), fabs(xnn-xn) > tolerance );
                    },
                    [](auto xn, auto xnn){ return xnn; },
                    start, 0.0 );
}

void NewtonTest()
{
    auto f = [](double x){ return x*x - 612.; };
    auto df = [](double x){ return 2. * x; };

    auto res = NewtonIterator(f, df, 10., 1e-8);
    printf("Neston iterator result: %16.16f\n", res);
}

template<bool b> struct Bool
{
	using type = bool;
    static const type value = b;
    constexpr operator bool() const { return value; }
};

template<typename T> struct Exist;
template<>           struct Exist<Bool<true>>{ using result = Bool<true>; };

template<int n> struct Int
{
	using type = int;
    static const type value = n;
    constexpr operator int() const { return value; }
    auto operator++()    const { return Int<(n+1)>(); }
    auto operator++(int) const { return Int<(n+1)>(); }
    auto operator--()    const { return Int<(n-1)>(); }
    auto operator--(int) const { return Int<(n-1)>(); }
};

template<int n1, int n2> auto operator+(Int<n1>, Int<n2>){ return Int<n1 + n2>(); }
template<int n1, int n2> auto operator-(Int<n1>, Int<n2>){ return Int<n1 - n2>(); }
template<int n1, int n2> auto operator*(Int<n1>, Int<n2>){ return Int<n1 * n2>(); }
template<int n1, int n2> auto operator/(Int<n1>, Int<n2>){ return Int<n1 / n2>(); }
template<int n1, int n2> auto operator==(Int<n1>, Int<n2>){ return Bool<(n1 == n2)>(); }

template<int64_t n, int64_t d> struct Rat{ static const int64_t nom = n; static const int64_t den = d; operator double() const { return (double)n / (double)d; } };

template<int64_t n1, int64_t n2, int64_t d1, int64_t d2> auto operator+( Rat<n1, d1>, Rat<n2, d2> ){ using res = std::ratio_add< std::ratio<n1, d1>, std::ratio<n2, d2> >; return Rat< res::num, res::den >(); }
template<int64_t n1, int64_t n2, int64_t d1, int64_t d2> auto operator-( Rat<n1, d1>, Rat<n2, d2> ){ using res = std::ratio_subtract< std::ratio<n1, d1>, std::ratio<n2, d2> >; return Rat< res::num, res::den >(); }
template<int64_t n1, int64_t n2, int64_t d1, int64_t d2> auto operator*( Rat<n1, d1>, Rat<n2, d2> ){ using res = std::ratio_multiply< std::ratio<n1, d1>, std::ratio<n2, d2> >; return Rat< res::num, res::den >(); }
template<int64_t n1, int64_t n2, int64_t d1, int64_t d2> auto operator/( Rat<n1, d1>, Rat<n2, d2> ){ using res = std::ratio_divide< std::ratio<n1, d1>, std::ratio<n2, d2> >; return Rat< res::num, res::den >(); }


//Product type:
template<typename N, typename T> struct elem       { T val; elem():val(T()){} elem( T const& v ):val(v){} elem(T && v) :val(std::forward<T>(v)){} };
template<typename N, typename T> struct elem<N, T&>{ T val;                   elem( T const& v ):val(v){}                               };

#define el(n, T) elem< decltype(n), T >

template<typename... Ts> struct List
{
};

template<typename Nk, typename Nn, typename T> struct ProductType_;
template<typename Nk>                          struct ProductType_<Nk, Int<0>, List<>>{};

template<typename Nk, typename Nn, typename T0, typename... T>
struct ProductType_<Nk, Nn, List<T0, T...>> : elem< decltype(--Nk()), T0>, ProductType_< decltype(Nk()++), Nn, List<T...>>
{
    using ProductType_< decltype(Nk()++), Nn, List<T...>>::operator[];
    ProductType_(){}
    ProductType_(ProductType_<Nk, Nn, List<T0, T...>>const& p): elem<decltype(--Nk()), T0>(p[--Nk()]), ProductType_< decltype(Nk()++), Nn, List<T...>>( static_cast<ProductType_< decltype(Nk()++), Nn, List<T...>>const&>(p)){}
    ProductType_(ProductType_<Nk, Nn, List<T0, T...>>&&     p): elem<decltype(--Nk()), T0>(p[--Nk()]), ProductType_< decltype(Nk()++), Nn, List<T...>>( static_cast<ProductType_< decltype(Nk()++), Nn, List<T...>>&&    >(p)){}

    auto& operator= (ProductType_ const& cpy)
    {
        (static_cast<elem< decltype(--Nk()), T0>*>(this))->val = (static_cast<const elem< decltype(--Nk()), T0>*>(&cpy))->val;
        *(static_cast<ProductType_< decltype(Nk()++), Nn, List<T...>>*>(this)) = *(static_cast<const ProductType_< decltype(Nk()++), Nn, List<T...>>*>(&cpy));
        return *this;
    }

    template<typename T0c, typename... Tsc> ProductType_(T0c && t, Tsc &&... ts) : elem< decltype(--Nk()), T0>( std::forward<T0c>(t) ), ProductType_< decltype(Nk()++), Nn, List<T...>>( std::forward<Tsc>(ts)... ) {}

    T0&       operator[]( decltype(--Nk()) idx)     { return (static_cast<      elem<decltype(--Nk()), T0>*>(this))->val; }
    T0 const& operator[]( decltype(--Nk()) idx)const{ return (static_cast<const elem<decltype(--Nk()), T0>*>(this))->val; }
};

template<typename Nn, typename T> struct ProductType_<Nn, Nn, List<T> > : elem< decltype(--Nn()), T>
{
    ProductType_(){}
    ProductType_(ProductType_<Nn, Nn, List<T>>const& cpy):elem< decltype(--Nn()), T>(cpy[--Nn()]){}
    ProductType_(ProductType_<Nn, Nn, List<T>>&&     cpy):elem< decltype(--Nn()), T>(cpy[--Nn()]){}

    auto& operator= ( ProductType_<Nn, Nn, List<T>> const& cpy){ (static_cast<elem< decltype(--Nn()), T>*>(this))->val = (static_cast<const elem< decltype(--Nn()), T>*>(&cpy))->val; return *this; }

    template<typename Tc> ProductType_(Tc && t):elem< decltype(--Nn()), T>(std::forward<Tc>(t)){}
	T&        operator[]( decltype(--Nn()) idx)     { return (static_cast<      elem< decltype(--Nn()), T>*>(this))->val; }
    T const & operator[]( decltype(--Nn()) idx)const{ return (static_cast<const elem< decltype(--Nn()), T>*>(this))->val; }
};

template<typename... T> struct Tuple
{
	using size = Int<sizeof...(T)>;
    ProductType_<Int<1>, Int<sizeof...(T)>, List<T...>> p;

    Tuple(){}
    Tuple(Tuple<T...> const& pin):p(          pin.p) {}
    Tuple(Tuple<T...> &&     pin):p(std::move(pin.p)){}

    template<typename... Tin> Tuple( int dummy, Tin&&... ts):p( std::forward<Tin>(ts)... ){}

    auto& operator= (Tuple<T...> const& cpy){ p = cpy.p; return *this; }

    template<int i> auto &      operator[]( Int<i> idx)     { return p[idx]; }
    template<int i> auto const& operator[]( Int<i> idx)const{ return p[idx]; }
};

template<typename T>     struct IsTuple;
template<typename... Ts> struct IsTuple< Tuple<Ts...> >{ using result = void; };
template<typename... Ts> struct IsTuple< const Tuple<Ts...> >{ using result = void; };

template<typename T, typename R = typename IsTuple<std::remove_reference_t<T>>::result> auto size( T&& tuple ){ return typename std::remove_reference_t<T>::size(); }

template<typename... Ts>
auto makeTuple( Ts&&... ts){ return Tuple<std::remove_reference_t<Ts>...>( 0, std::forward<Ts>(ts)... ); }

//Expand
template<typename Ni, typename Nn> struct expand_
{
	template<typename F, typename Tup, typename... A>
	decltype(auto) operator()( F&& f, Tup&& t, A&&... a){ return expand_< decltype(++Ni()), Nn>()( std::forward<F>(f), std::forward<Tup>(t), std::forward<A>(a)..., t[--Ni()] ); }
};

template<typename Nn> struct expand_<Nn, Nn>
{
	template<typename F, typename Tup, typename... A>
	decltype(auto) operator()( F&& f, Tup&& t, A&&... a){ return f( std::forward<A>(a)..., t[--Nn()] ); }
};

template<typename F, typename Tup>
decltype(auto) expand(F&& f, Tup&& t){ return expand_<Int<1>, decltype(size(t))>()(std::forward<F>(f), std::forward<Tup>(t)); }

//Map
template<typename F, typename T, typename R = typename IsTuple<std::remove_reference_t<T>>::result>
auto Map( F&& f, T&& t)
{
	return expand( [&](auto... e){ return makeTuple( f( std::forward<decltype(e)>(e))... ); }, std::forward<T>(t) );
}

//FOLD
template<typename F> struct Foldl_inner_lambda
{
    F f;
	Foldl_inner_lambda(F&& ff):f(std::move(ff)){}
	Foldl_inner_lambda(F const& ff):f(ff){}

    template<typename T1>                                          auto operator()(T1&& t1)         { return std::forward<T1>(t1); }
    template<typename T1, typename T2>                             auto operator()(T1&& t1, T2&& t2){ return f(std::forward<T1>(t1), std::forward<T2>(t2)); }
    template<typename T1, typename T2, typename T3, typename... T> auto operator()(T1&& t1, T2&& t2, T3&& t3, T&&... t)
    {
		return this->operator()( f(std::forward<T1>(t1), std::forward<T2>(t2)), std::forward<T3>(t3), std::forward<T>(t)... );
    }
};

template<typename F, typename T>
auto Foldl(F&& f, T&& t){ return expand( Foldl_inner_lambda<std::remove_cv_t<std::remove_reference_t<F>>>(std::forward<F>(f)), std::forward<T>(t) ); }

//CONCATENATE
template<typename     T, typename   Pck> struct Concat;
template<typename     T, typename... Ts> struct Concat<T, Tuple<Ts...>>           { using result = Tuple<T, Ts...>; };
template<typename... Ts, typename     T> struct Concat<Tuple<Ts...>, T>           { using result = Tuple<Ts..., T>; };
template<typename... As, typename... Bs> struct Concat<Tuple<As...>, Tuple<Bs...>>{ using result = Tuple<As..., Bs...>; };

template<typename V, typename T>
auto append( T&& t, V&& val ){ return expand( [&](auto&&... e){ return makeTuple( std::forward<decltype(e)>(e)..., std::forward<V>(val) ); }, std::forward<T>(t) ); }

//generate nat sequence:
auto genseq( Int<1> ){ return makeTuple( Int<0>() ); }
template<int n> auto genseq(Int<n> i){ auto k = --i; return append( genseq( k ), k ); }

//Zip
template<typename F, typename T1, typename T2, typename R = typename Exist<decltype(size(std::declval<T1>()) == size(std::declval<T2>()) )>::result>
auto Zip( F&& f, T1&& t1, T2&& t2)
{
	return expand( [&](auto... ns){ return makeTuple( f( t1[ns], t2[ns] )... ); }, genseq(size(t1)) );
}

#include <iostream>
void TupleTest()
{
	auto t = makeTuple( Int<5>(), 8, Int<7>() );
	auto r = Map( [](auto x){ return x*x; }, t);

	printf("\n%i  %i  %i\n", t[Int<0>()].value, t[Int<1>()], t[Int<2>()].value);
	printf("\n%i  %i  %i\n", r[Int<0>()].value, r[Int<1>()], r[Int<2>()].value);

	auto t2 = makeTuple( Int<5>(), Int<1>(), Int<7>() );
	auto r2 = Foldl( [](auto x, auto y){ return x+y; }, t2);

	printf("\n%i  %i  %i\n", t2[Int<0>()].value, t2[Int<1>()].value, t2[Int<2>()].value);
	printf("\n%i\n", r2.value );

	auto t3 = makeTuple( Int<5>(), Int<1>(), Int<7>() );
	auto t4 = makeTuple( Int<2>(), Int<7>(), Int<9>() );
	auto r3 = Zip( [](auto x, auto y){ return x+y; }, t3, t4);

	printf("\n%i  %i  %i\n", t3[Int<0>()].value, t3[Int<1>()].value, t3[Int<2>()].value);
	printf("\n%i  %i  %i\n", t4[Int<0>()].value, t4[Int<1>()].value, t4[Int<2>()].value);
	printf("\n%i  %i  %i\n", r3[Int<0>()].value, r3[Int<1>()].value, r3[Int<2>()].value);
}

auto Add = [](auto&& x, auto&& y){ return std::forward<decltype(x)>(x)+std::forward<decltype(y)>(y); };
auto Sub = [](auto&& x, auto&& y){ return std::forward<decltype(x)>(x)-std::forward<decltype(y)>(y); };
//auto Mul = [](auto&& x, auto&& y){ return std::forward<decltype(x)>(x)*std::forward<decltype(y)>(y); };
auto SclMul = [](auto&& scl){ return [&](auto&& x){ return scl*std::forward<decltype(x)>(x); }; };
template<typename T>
auto sq( T&& x ){ return x*x; }

template<typename As, typename Bs, typename Bps, typename Cs>
struct RungeKuttaStepper
{
	As  as;
	Bs  bs;
    Bps bps;
	Cs  cs;

	template<typename TTime, typename TStep, typename TState, typename TRHS>
	auto eval_kn( Int<0>, TTime&& time, TStep&& h, TState&& state, TRHS&& rhs ) const
	{
		return makeTuple(rhs(time, std::forward<TState>(state)));
	}

	template<typename N, typename TTime, typename TStep, typename TState, typename TRHS>
	auto eval_kn( N&& n, TTime&& time, TStep&& h, TState&& state, TRHS&& rhs ) const
	{
		auto ks = eval_kn(--n, time, h, std::forward<TState>(state), std::forward<TRHS>(rhs) );
		auto StateAdder = [](auto&& s1, auto&& s2){ return Zip(Add, s1, s2); };
		auto StateScaler = [&](auto&& scl, auto&& s){ return Map(SclMul(scl), s); };
		auto kc = rhs( time + cs[n] * h,
			           Zip(Add, state, Map(SclMul(h), Foldl(StateAdder, Zip(StateScaler, as[n], ks) ) ) ) );
		return append(ks, kc);
	}

	template<typename TTime, typename TStep, typename TState, typename TRHS>
	auto operator()( TTime&& time, TStep&& h, TState&& state, TRHS&& rhs ) const
	{
		auto ks = eval_kn( --size(bs), time, h, state, rhs );
		auto StateAdder = [](auto&& s1, auto&& s2){ return Zip(Add, s1, s2); };
		auto StateScaler = [&](auto&& scl, auto&& s){ return Map(SclMul(scl), s); };
		return makeTuple( state, Zip( Add, state, Map( SclMul(h), Foldl( StateAdder, Zip(StateScaler, bs, ks)))),
                                 Foldl( StateAdder, Zip(StateScaler, Zip(Sub, bs, bps), ks)) );//Error state vector
	}
};

template<typename UF, typename FF, typename Z, typename SH, typename S>
auto mutu( UF&& uf, FF&& ff, Z&& zero, SH&& sh, S&& seed )
{
    auto ufs = uf( std::forward<SH>(sh), std::forward<S>(seed) );
    auto fs = ff( std::forward<Z>(zero), ufs.l );
    while( fs )
    {
        ufs = uf( fs.t.r, ufs.r );
        fs = ff( fs.t.l, ufs.l );
    }
    return makeTuple(fs.t.l, fs.t.r, ufs.r);
}

template<typename TStepper, typename TTime, typename TState, typename TRHS>
auto RungeKuttaRepeater( TStepper&& stepper, TTime&& time, TTime&& h, TState&& state, TRHS&& rhs )
{
    struct FoldState
    {
        double atol, rtol, last_error;
        bool reject;
    };
    FoldState fs0{1e-7, 1e-7, 1e-4, false};

    struct UnfoldState{} ufs0;

    return mutu(
        [&]( auto&& ths, UnfoldState& ufs )
        {
            return makePair( makeTuple(ths[Int<0>()], ths[Int<1>()], stepper( ths[Int<0>()], ths[Int<1>()], ths[Int<2>()], rhs )), ufs );
        },
        []( FoldState& fs, auto&& thsds )
        {
            auto state0 = thsds[Int<2>()][Int<0>()];
            auto state1 = thsds[Int<2>()][Int<1>()];
            auto& dstate = thsds[Int<2>()][Int<2>()];
            auto err_pre = Foldl( Add, Map( [=](auto const& p){ return sq(p.r / (fs.atol + fabs(p.l)*fs.rtol)); }, Zip( [](auto const& l, auto const& r){ return makePair(l, r); }, state1, dstate ) ) );
            auto err = sqrt(err_pre / (int)size(state0));

            auto& t = thsds[Int<0>()];
            auto& hc = thsds[Int<1>()];

            auto hnext = 0.95 * fabs(hc) * pow( fabs(fs.last_error), 0.4/5.0) * pow(fabs(err), -0.7/5.0);
            bool end = (err <= 1.0);
            fs.last_error = err;

            return escaper( makePair(fs, makeTuple((end ? t+hc : t), hnext, (end ? state1 : state0) )), !end );
        },
        fs0, makeTuple( time, h, std::forward<TState>(state) ), ufs0 );
}

template<typename TStepper, typename TTime, typename TState, typename TRHS, typename FFoldFunc, typename FoldState, typename FEndCondition>
auto ODEDriver(TStepper&& stepper, TTime&& t0, TState&& initial_state, TRHS&& rhs, FFoldFunc&& ff, FoldState&& foldstate, FEndCondition&& ec)
{
    FoldState foldstate0 = std::forward<FoldState>(foldstate);
    struct UnfoldState{} ufs0;

    return mutu(
        [&](auto && ths, auto&& ufs)
        {
            auto tmp = RungeKuttaRepeater(stepper, ths[Int<0>()], ths[Int<1>()], ths[Int<2>()], rhs)[Int<1>()];
            return makePair( tmp, ufs );
        },
        [&](FoldState& fs, auto && ths)
        {
            auto&& tmp = ff(fs, ths);
            return escaper( makePair(fs, tmp), ec(tmp) );
        },
        foldstate0, makeTuple(t0, 1e-4, initial_state), ufs0
        );
}

struct LVState
{
    double r, w;
    LVState( double r0, double w0 ):r(r0), w(w0){}
    LVState( Tuple<double, double> const& t ):r(t[Int<0>()]), w(t[Int<1>()]){}
    double const& operator[]( Int<0> )const{ return r; }
    double const& operator[]( Int<1> )const{ return w; }
};

auto size( LVState const& ){ return Int<2>(); }

template<typename F>
LVState Map(F&& f, LVState const& lvs){ return LVState{ f(lvs.r), f(lvs.w) }; }

template<typename F>
LVState Zip(F&& f, LVState const& lvs1, LVState const& lvs2){ return LVState{ f(lvs1.r, lvs2.r), f(lvs1.w, lvs2.w) }; }

#include <memory>

void main23()
{
    //Dormand-Prince 5 stepper:
    //https://en.wikipedia.org/wiki/Dormand–Prince_method

    using As = Tuple<	Tuple< Rat<0, 1> >,
					    Tuple< Rat<1, 5> >,
                        Tuple< Rat<3, 40>, Rat<9, 40> >,
                        Tuple< Rat<44, 45>, Rat<-56, 15>, Rat<32, 9> >,
                        Tuple< Rat<19372, 6561>, Rat<-25360, 2187>, Rat<64448, 6561>, Rat<-212, 729> >,
                        Tuple< Rat<9017, 3168>, Rat<-355, 33>, Rat<46732, 5247>, Rat<49, 176>, Rat<-5103, 18656> >,
                        Tuple< Rat<35, 384>, Rat<0, 1>, Rat<500, 1113>, Rat<125, 192>, Rat<-2187, 6784>, Rat<11, 84> >
                    >;

	using Bs = Tuple<	Rat<35, 384>,
						Rat<0, 1>,
                        Rat<500, 1113>,
                        Rat<125, 192>,
                        Rat<-2187, 6784>,
                        Rat<11, 84>,
                        Rat<0, 1>
                    >;
    using Bps = Tuple<	Rat<5179, 57600>,
						Rat<0, 1>,
                        Rat<7571, 16695>,
                        Rat<393, 640>,
                        Rat<-92097, 339200>,
                        Rat<187, 2100>,
                        Rat<1, 40>
                    >;

	using Cs = Tuple<	Rat<0, 1>,
						Rat<1, 5>,
                        Rat<3, 10>,
                        Rat<4, 5>,
                        Rat<8, 9>,
                        Rat<1, 1>,
                        Rat<1, 1>
                    >;

    RungeKuttaStepper< As, Bs, Bps, Cs > dps;

    //Lotka-Volterra
    double a = 0.5, b = 0.1, c = 0.8, d = 0.3;

    double t0 = 0.0;
    LVState s0{ 10.0, 1.0 };
    auto rhs = [=](double t, LVState const& s){ return makeTuple( a*s.r - b*s.r*s.w, d*s.r*s.w - c*s.w ); };
    auto fendcond = [](auto&& ths){ return ths[Int<0>()] < 120.0; };

    std::ofstream file("out.txt");

    struct Logger
    {
        std::ofstream* pfile;
    };

    auto foldfunc = []( Logger& log, auto&& ths )
    {
        auto& t = ths[ Int<0>() ];
        auto& h = ths[ Int<1>() ];
        auto& s = ths[ Int<2>() ];
        *log.pfile << t << "   " << h << "   " << s[Int<0>()] << "   " << s[Int<1>()] << "\n";
        return ths;
    };

    Logger logger{&file};

    auto res = ODEDriver(dps, t0, makeTuple(10.0, 1.0), rhs, foldfunc, logger, fendcond );
}
